print("import: 'thinc'")
import thinc

print("import: 'thinc.api'")
import thinc.api

print("import: 'thinc.backends'")
import thinc.backends

print("import: 'thinc.extra'")
import thinc.extra

print("import: 'thinc.layers'")
import thinc.layers

print("import: 'thinc.shims'")
import thinc.shims

